import functools
import inspect
import warnings
from typing import Callable, Generic, Optional, TypeVar, overload

from ntcore import NetworkTableInstance, Value

T = TypeVar("T")
V = TypeVar("V")


class tunable(Generic[V]):
    """
    This allows you to define simple properties that allow you to easily
    communicate with other programs via NetworkTables.

    The following example will define a NetworkTable variable at
    ``/components/my_component/foo``::

        class MyRobot(magicbot.MagicRobot):

            my_component: MyComponent

        ...

        from magicbot import tunable

        class MyComponent:

            # define the tunable property
            foo = tunable(True)

            def execute(self):

                # set the variable
                self.foo = True

                # get the variable
                foo = self.foo

    The key of the NetworkTables variable will vary based on what kind of
    object the decorated method belongs to:

    * A component: ``/components/COMPONENTNAME/VARNAME``
    * An autonomous mode: ``/autonomous/MODENAME/VARNAME``
    * Your main robot class: ``/robot/VARNAME``

    .. note:: When executing unit tests on objects that create tunables,
              you will want to use setup_tunables to set the object up.
              In normal usage, MagicRobot does this for you, so you don't
              have to do anything special.
    """

    # the way this works is we use a special class to indicate that it
    # is a tunable, and MagicRobot adds _ntattr and _global_table variables
    # to the class property

    # The tricky bit is that you need to do late binding on these, because
    # the networktables key is not known when the object is created. Instead,
    # the name of the key is related to the name of the variable name in the
    # robot class

    __slots__ = (
        "_ntdefault",
        "_ntsubtable",
        "_ntwritedefault",
        # "__doc__",
        "_mkv",
        "_nt",
    )

    def __init__(
        self,
        default: V,
        *,
        writeDefault: bool = True,
        subtable: Optional[str] = None,
        doc=None,
    ) -> None:
        if doc is not None:
            warnings.warn("tunable no longer uses the doc argument", stacklevel=2)

        self._ntdefault = default
        self._ntsubtable = subtable
        self._ntwritedefault = writeDefault
        d = Value.makeValue(default)
        self._mkv = Value.getFactoryByType(d.type())
        # self.__doc__ = doc

    @overload
    def __get__(self, instance: None, owner=None) -> "tunable[V]":
        ...

    @overload
    def __get__(self, instance, owner=None) -> V:
        ...

    def __get__(self, instance, owner=None):
        if instance is not None:
            return instance._tunables[self].value
        return self

    def __set__(self, instance, value: V) -> None:
        instance._tunables[self].setValue(self._mkv(value))


def setup_tunables(component, cname: str, prefix: Optional[str] = "components") -> None:
    """
    Connects the tunables on an object to NetworkTables.

    :param component:   Component object
    :param cname:       Name of component
    :param prefix:      Prefix to use, or no prefix if None

    .. note:: This is not needed in normal use, only useful
              for testing
    """

    cls = component.__class__

    if prefix is None:
        prefix = "/%s" % cname
    else:
        prefix = "/%s/%s" % (prefix, cname)

    NetworkTables = NetworkTableInstance.getDefault()

    tunables = {}

    for n in dir(cls):
        if n.startswith("_"):
            continue

        prop = getattr(cls, n)
        if not isinstance(prop, tunable):
            continue

        if prop._ntsubtable:
            key = "%s/%s/%s" % (prefix, prop._ntsubtable, n)
        else:
            key = "%s/%s" % (prefix, n)

        ntvalue = NetworkTables.getEntry(key)
        if prop._ntwritedefault:
            ntvalue.setValue(prop._ntdefault)
        else:
            ntvalue.setDefaultValue(prop._ntdefault)
        tunables[prop] = ntvalue

    component._tunables = tunables


@overload
def feedback(f: Callable[[T], V]) -> Callable[[T], V]:
    ...


@overload
def feedback(*, key: str) -> Callable[[Callable[[T], V]], Callable[[T], V]]:
    ...


def feedback(f=None, *, key: Optional[str] = None) -> Callable:
    """
    This decorator allows you to create NetworkTables values that are
    automatically updated with the return value of a method.

    ``key`` is an optional parameter, and if it is not supplied,
    the key will default to the method name with a leading ``get_`` removed.
    If the method does not start with ``get_``, the key will be the full
    name of the method.

    The key of the NetworkTables value will vary based on what kind of
    object the decorated method belongs to:

    * A component: ``/components/COMPONENTNAME/VARNAME``
    * Your main robot class: ``/robot/VARNAME``

    The NetworkTables value will be auto-updated in all modes.

    .. warning:: The function should only act as a getter, and must not
                 take any arguments (other than self).

    Example::

        from magicbot import feedback

        class MyComponent:
            navx: ...

            @feedback
            def get_angle(self):
                return self.navx.getYaw()

        class MyRobot(magicbot.MagicRobot):
            my_component: MyComponent

            ...

    In this example, the NetworkTable key is stored at
    ``/components/my_component/angle``.

    .. seealso:: :class:`~wpilib.LiveWindow` may suit your needs,
                 especially if you wish to monitor WPILib objects.

    .. versionadded:: 2018.1.0
    """
    if f is None:
        return functools.partial(feedback, key=key)

    if not callable(f):
        raise TypeError(f"Illegal use of feedback decorator on non-callable {f!r}")
    sig = inspect.signature(f)
    name = f.__name__

    if len(sig.parameters) != 1:
        raise ValueError(
            f"{name} may not take arguments other than 'self' (must be a simple getter method)"
        )

    # Set attributes to be checked during injection
    f._magic_feedback = True
    f._magic_feedback_key = key

    return f


def collect_feedbacks(component, cname: str, prefix: Optional[str] = "components"):
    """
    Finds all methods decorated with :func:`feedback` on an object
    and returns a list of 2-tuples (method, NetworkTables entry).

    .. note:: This isn't useful for normal use.
    """
    if prefix is None:
        prefix = "/%s" % cname
    else:
        prefix = "/%s/%s" % (prefix, cname)

    nt = NetworkTableInstance.getDefault().getTable(prefix)
    feedbacks = []

    for name, method in inspect.getmembers(component, inspect.ismethod):
        if getattr(method, "_magic_feedback", False):
            key = method._magic_feedback_key
            if key is None:
                if name.startswith("get_"):
                    key = name[4:]
                else:
                    key = name

            entry = nt.getEntry(key)
            feedbacks.append((method, entry))

    return feedbacks



# TODO: remove this once WPILib is public, and use the real thing

import sys

import pytest
from unittest.mock import MagicMock


def pytest_runtest_setup():
    pass


def pytest_runtest_teardown():
    pass


@pytest.fixture(scope="function")
def wpimock(monkeypatch):
    mock = MagicMock(name="wpimock")
    monkeypatch.setitem(sys.modules, "wpilib", mock)
    return mock


@pytest.fixture(scope="function")
def wpitime():
    import hal.simulation

    class FakeTime:
        def step(self, seconds):
            delta = int(seconds * 1000000)
            hal.simulation.stepTimingAsync(delta)

    hal.simulation.pauseTiming()
    hal.simulation.restartTiming()

    yield FakeTime()

    hal.simulation.resumeTiming()


@pytest.fixture(scope="function")
def hal(wpitime):
    import hal.simulation

    yield

    # Reset the HAL handles
    hal.simulation.resetGlobalHandles()

    # Reset the HAL data
    hal.simulation.resetAllSimData()



def test_sensor(wpimock):
    from robotpy_ext.common_drivers.pressure_sensors import REVAnalogPressureSensor

    sensor = REVAnalogPressureSensor(3, 3.3)
    sensor.sensor.getAverageVoltage.return_value = 2.0
    assert int(sensor.pressure) == 126


def test_calibration(wpimock):
    from robotpy_ext.common_drivers.pressure_sensors import REVAnalogPressureSensor

    sensor = REVAnalogPressureSensor(3, 3.3)
    sensor.sensor.getAverageVoltage.return_value = 2.0
    sensor.calibrate(50)
    assert int(sensor.pressure) == 50



from magicbot.state_machine import (
    default_state,
    state,
    timed_state,
    AutonomousStateMachine,
    StateMachine,
    IllegalCallError,
    NoFirstStateError,
    MultipleFirstStatesError,
    MultipleDefaultStatesError,
    InvalidStateName,
)

from magicbot.magic_tunable import setup_tunables

import pytest


def test_no_timed_state_duration():
    with pytest.raises(TypeError):

        class _TM(StateMachine):
            @timed_state()
            def tmp(self):
                pass


def test_no_start_state():
    class _TM(StateMachine):
        pass

    with pytest.raises(NoFirstStateError):
        _TM()


def test_multiple_first_states():
    class _TM(StateMachine):
        @state(first=True)
        def tmp1(self):
            pass

        @state(first=True)
        def tmp2(self):
            pass

    with pytest.raises(MultipleFirstStatesError):
        _TM()


def test_sm(wpitime):
    class _TM(StateMachine):
        def __init__(self):
            self.executed = []

        def some_fn(self):
            self.executed.append("sf")

        @state(first=True)
        def first_state(self):
            self.executed.append(1)
            self.next_state("second_state")

        @timed_state(duration=1, next_state="third_state")
        def second_state(self):
            self.executed.append(2)

        @state
        def third_state(self):
            self.executed.append(3)

    sm = _TM()
    setup_tunables(sm, "test_sm")
    sm.some_fn()

    # should not be able to directly call
    with pytest.raises(IllegalCallError):
        sm.first_state()

    assert sm.current_state == ""
    assert not sm.is_executing

    sm.engage()
    assert sm.current_state == "first_state"
    assert not sm.is_executing

    sm.execute()
    assert sm.current_state == "second_state"
    assert sm.is_executing

    # should not change
    sm.engage()
    assert sm.current_state == "second_state"
    assert sm.is_executing

    sm.execute()
    assert sm.current_state == "second_state"
    assert sm.is_executing

    wpitime.step(1.5)
    sm.engage()
    sm.execute()
    assert sm.current_state == "third_state"
    assert sm.is_executing

    sm.engage()
    sm.execute()
    assert sm.current_state == "third_state"
    assert sm.is_executing

    # should be done
    sm.done()
    assert sm.current_state == ""
    assert not sm.is_executing

    # should be able to start directly at second state
    sm.engage(initial_state="second_state")
    sm.execute()
    assert sm.current_state == "second_state"
    assert sm.is_executing

    wpitime.step(1.5)
    sm.engage()
    sm.execute()
    assert sm.current_state == "third_state"
    assert sm.is_executing

    # test force
    sm.engage()
    sm.execute()
    assert sm.current_state == "third_state"
    assert sm.is_executing

    sm.engage(force=True)
    assert sm.current_state == "first_state"
    assert sm.is_executing

    sm.execute()
    sm.execute()
    assert not sm.is_executing
    assert sm.current_state == ""

    assert sm.executed == ["sf", 1, 2, 3, 3, 2, 3, 3, 1]


def test_sm_inheritance():
    class _TM1(StateMachine):
        @state
        def second_state(self):
            self.done()

    class _TM2(_TM1):
        @state(first=True)
        def first_state(self):
            self.next_state("second_state")

    sm = _TM2()
    setup_tunables(sm, "test_sm_inheritance")
    sm.engage()
    assert sm.current_state == "first_state"

    sm.execute()
    assert sm.current_state == "second_state"

    sm.execute()
    assert sm.current_state == ""


def test_must_finish(wpitime):
    class _TM(StateMachine):
        def __init__(self):
            self.executed = []

        @state(first=True)
        def ordinary1(self):
            self.next_state("ordinary2")
            self.executed.append(1)

        @state
        def ordinary2(self):
            self.next_state("must_finish")
            self.executed.append(2)

        @state(must_finish=True)
        def must_finish(self):
            self.executed.append("mf")

        @state
        def ordinary3(self):
            self.executed.append(3)
            self.next_state_now("timed_must_finish")

        @timed_state(duration=1, must_finish=True)
        def timed_must_finish(self):
            self.executed.append("tmf")

    sm = _TM()
    setup_tunables(sm, "test_must_finish")

    sm.engage()
    sm.execute()
    sm.execute()

    assert sm.current_state == ""
    assert not sm.is_executing

    sm.engage()
    sm.execute()
    sm.engage()
    sm.execute()
    sm.execute()
    sm.execute()

    assert sm.current_state == "must_finish"
    assert sm.is_executing

    sm.next_state("ordinary3")
    sm.engage()
    sm.execute()

    assert sm.current_state == "timed_must_finish"

    sm.execute()
    assert sm.is_executing
    assert sm.current_state == "timed_must_finish"

    for _ in range(7):
        wpitime.step(0.1)

        sm.execute()
        assert sm.is_executing
        assert sm.current_state == "timed_must_finish"

    wpitime.step(1)
    sm.execute()
    assert not sm.is_executing

    assert sm.executed == [1, 1, 2, "mf", "mf", 3] + ["tmf"] * 9


def test_autonomous_sm():
    class _TM(AutonomousStateMachine):

        i = 0
        VERBOSE_LOGGING = False

        @state(first=True)
        def something(self):
            self.i += 1
            if self.i == 6:
                self.done()

    sm = _TM()
    setup_tunables(sm, "test_autonomous_sm")

    sm.on_enable()

    for _ in range(5):
        sm.on_iteration(None)
        assert sm.is_executing

    sm.on_iteration(None)
    assert not sm.is_executing

    for _ in range(5):
        sm.on_iteration(None)
        assert not sm.is_executing

    assert sm.i == 6


def test_autonomous_sm_end_timed_state(wpitime):
    class _TM(AutonomousStateMachine):

        i = 0
        j = 0
        VERBOSE_LOGGING = False

        @state(first=True)
        def something(self):
            self.i += 1
            if self.i == 3:
                self.next_state("timed")

        @timed_state(duration=1)
        def timed(self):
            self.j += 1

    sm = _TM()
    setup_tunables(sm, "test_autonomous_sm_end_timed_state")

    sm.on_enable()

    for _ in range(5):
        wpitime.step(0.7)
        sm.on_iteration(None)
        assert sm.is_executing

    for _ in range(5):
        wpitime.step(0.7)
        sm.on_iteration(None)
        assert not sm.is_executing

    assert sm.i == 3
    assert sm.j == 2


def test_next_fn():
    class _TM(StateMachine):
        @state(first=True)
        def first_state(self):
            self.next_state(self.second_state)

        @state
        def second_state(self):
            self.done()

    sm = _TM()
    setup_tunables(sm, "test_next_fn")
    sm.engage()
    assert sm.current_state == "first_state"

    sm.execute()
    assert sm.current_state == "second_state"

    sm.engage()
    sm.execute()
    assert sm.current_state == ""


def test_next_fn2(wpitime):
    class _TM(StateMachine):
        @state
        def second_state(self):
            pass

        @timed_state(first=True, duration=0.1, next_state=second_state)
        def first_state(self):
            pass

    sm = _TM()
    setup_tunables(sm, "test_next_fn2")
    sm.engage()
    sm.execute()
    assert sm.current_state == "first_state"
    assert sm.is_executing

    wpitime.step(0.5)

    sm.engage()
    sm.execute()
    assert sm.current_state == "second_state"
    assert sm.is_executing

    sm.execute()
    assert sm.current_state == ""
    assert not sm.is_executing


def test_mixup():
    from robotpy_ext.autonomous import state as _ext_state
    from robotpy_ext.autonomous import timed_state as _ext_timed_state

    with pytest.raises(RuntimeError) as exc_info:

        class _SM1(StateMachine):
            @_ext_state(first=True)
            def the_state(self):
                pass

    assert isinstance(exc_info.value.__cause__, TypeError)

    with pytest.raises(RuntimeError) as exc_info:

        class _SM2(StateMachine):
            @_ext_timed_state(first=True, duration=1)
            def the_state(self):
                pass

    assert isinstance(exc_info.value.__cause__, TypeError)


def test_forbidden_state_names():
    with pytest.raises(InvalidStateName):

        class _SM(StateMachine):
            @state
            def done(self):
                pass


def test_mixins():
    class _SM1(StateMachine):
        @state
        def state1(self):
            pass

    class _SM2(StateMachine):
        @state
        def state2(self):
            pass

    class _SM(_SM1, _SM2):
        @state(first=True)
        def first_state(self):
            pass

    s = _SM()
    states = s._StateMachine__states

    assert "state1" in states
    assert "state2" in states
    assert "first_state" in states


def test_multiple_default_states():
    class _SM(StateMachine):
        @state(first=True)
        def state(self):
            pass

        @default_state
        def state1(self):
            pass

        @default_state
        def state2(self):
            pass

    with pytest.raises(MultipleDefaultStatesError):
        _SM()


def test_default_state_machine():
    class _SM(StateMachine):
        def __init__(self):
            self.didOne = None
            self.didDefault = None
            self.defaultInit = None
            self.didDone = None

        @state(first=True)
        def stateOne(self):
            self.didOne = True
            self.didDefault = False
            self.didDone = False

        @state
        def doneState(self):
            self.didOne = False
            self.didDefault = False
            self.didDone = True
            self.done()

        @default_state
        def defaultState(self, initial_call):
            self.didOne = False
            self.didDefault = True
            self.defaultInit = initial_call
            self.didDone = False

    sm = _SM()
    setup_tunables(sm, "test_default_state_machine")

    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == True
    assert sm.didDone == False

    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == False
    assert sm.didDone == False

    # do a thing
    sm.engage()
    sm.execute()
    assert sm.didOne == True
    assert sm.didDefault == False
    assert sm.didDone == False

    # should go back (test for initial)
    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == True
    assert sm.didDone == False

    # should happen again (no initial)
    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == False
    assert sm.didDone == False

    # do another thing
    sm.engage()
    sm.execute()
    assert sm.didOne == True
    assert sm.didDefault == False
    assert sm.didDone == False

    # should go back (test for initial)
    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == True
    assert sm.didDone == False

    # should happen again (no initial)
    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == False
    assert sm.didDone == False

    # enagage a state that will call done, check to see
    # if we come back
    sm.engage("doneState")
    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == False
    assert sm.defaultInit == False
    assert sm.didDone == True

    # should go back (test for initial)
    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == True
    assert sm.didDone == False

    # should happen again (no initial)
    sm.execute()
    assert sm.didOne == False
    assert sm.didDefault == True
    assert sm.defaultInit == False
    assert sm.didDone == False


def test_short_timed_state(wpitime):
    """
    Tests two things:
    - A timed state that expires before it executes
    - Ensures that the default state won't execute if the machine is always
      executing
    """

    class _SM(StateMachine):
        def __init__(self):
            self.executed = []

        @default_state
        def d(self):
            self.executed.append("d")

        @state(first=True)
        def a(self):
            self.executed.append("a")
            self.next_state("b")

        @timed_state(duration=0.01)
        def b(self):
            self.executed.append("b")

        def done(self):
            super().done()
            self.executed.append("d")

    sm = _SM()
    setup_tunables(sm, "test_short_timed_state")
    assert sm.current_state == ""
    assert not sm.is_executing

    for _ in [1, 2, 3, 4]:
        sm.engage()
        sm.execute()
        assert sm.current_state == "b"

        wpitime.step(0.02)

        sm.engage()
        sm.execute()
        assert sm.current_state == "b"

        wpitime.step(0.02)

    assert sm.executed == ["a", "b", "d", "a", "b", "d", "a", "b", "d", "a", "b"]



from robotpy_ext.control.toggle import Toggle
from robotpy_ext.misc.precise_delay import NotifierDelay


class FakeJoystick:
    def __init__(self):
        self._pressed = [False] * 2

    def getRawButton(self, num):
        return self._pressed[num]

    def press(self, num):
        self._pressed[num] = True

    def release(self, num):
        self._pressed[num] = False


def test_toggle():
    joystick = FakeJoystick()
    toggleButton = Toggle(joystick, 0)
    toggleButton2 = Toggle(joystick, 1)
    assert toggleButton.off
    joystick.press(0)
    assert toggleButton.on
    assert toggleButton2.off
    joystick.release(0)
    assert toggleButton.on
    joystick.press(0)
    assert toggleButton.off
    joystick.release(0)
    assert toggleButton.off
    joystick.press(1)
    assert toggleButton.off
    assert toggleButton2.on


def test_toggle_debounce():
    # TODO: use simulated time
    delay = NotifierDelay(0.5)
    joystick = FakeJoystick()
    toggleButton = Toggle(joystick, 1, 0.1)
    assert toggleButton.off
    joystick.press(1)
    assert toggleButton.on
    joystick.release(1)
    joystick.press(1)
    joystick.release(1)
    assert toggleButton.on
    delay.wait()
    assert toggleButton.on
    joystick.press(1)
    assert toggleButton.off



def get_class_annotations(cls):
    """Get variable annotations in a class, inheriting from superclasses."""
    result = {}
    for base in reversed(cls.__mro__):
        result.update(getattr(base, "__annotations__", {}))
    return result



import logging
import time


class PeriodicFilter:
    """
    Periodic Filter to help keep down clutter in the console.
    Simply add this filter to your logger and the logger will
    only print periodically.

    The logger will always print logging levels of WARNING or higher,
    unless given a different bypass level

    Example::

        class Component1:

            def setup(self):
                # Set period to 3 seconds, set bypass_level to WARN
                self.logger.addFilter(PeriodicFilter(3, bypass_level=logging.WARN))

            def execute(self):
                # This message will be printed once every three seconds
                self.logger.info('Component1 Executing')

                # This message will be printed out every loop
                self.logger.warn("Uh oh, this shouldn't have happened...")

    """

    def __init__(self, period, bypass_level=logging.WARN):
        """
        :param period: Wait period (in seconds) between logs
        :param bypass_level: Lowest logging level that the filter should not catch
        """

        self._period = period
        self._loggingLoop = True
        self._last_log = -period
        self._bypass_level = bypass_level

    def filter(self, record):
        """Performs filtering action for logger"""
        self._refresh_logger()
        return self._loggingLoop or record.levelno >= self._bypass_level

    def _refresh_logger(self):
        """Determine if the log wait period has passed"""
        now = time.monotonic()
        self._loggingLoop = False
        if now - self._last_log > self._period:
            self._loggingLoop = True
            self._last_log = now



import logging
from pyfrc.test_support.controller import TestController as PyfrcTestController
from ntcore.util import ChooserControl


autonomous_seconds = 15


def test_all_autonomous(control: PyfrcTestController):
    """
    This test runs all possible autonomous modes that can be selected
    by the autonomous switcher.

    This should work for most robots. If it doesn't work for yours,
    and it's not a code issue with your robot, please file a bug on
    github.
    """

    logger = logging.getLogger("test-all-autonomous")

    with control.run_robot():

        # Run disabled for a short period, chooser needs to be
        # initialized in robotInit
        control.step_timing(seconds=0.5, autonomous=True, enabled=False)

        # retrieve autonomous modes from chooser here
        chooser = ChooserControl("Autonomous Mode")
        choices = chooser.getChoices()
        if len(choices) == 0:
            return

        for choice in choices:

            chooser.setSelected(choice)
            logger.info(f"{'='*10} Testing '{choice}' {'='*10}")

            # Run disabled for a short period
            control.step_timing(seconds=0.5, autonomous=True, enabled=False)

            # Run enabled for 15 seconds
            control.step_timing(
                seconds=autonomous_seconds, autonomous=True, enabled=True
            )

            # Disabled for another short period
            control.step_timing(seconds=0.5, autonomous=True, enabled=False)
